@@ -24,7 +24,8 @@
  */
 
 #include <main.h>
-#define SYSTEM_FILE_NAME_MAX   40
+#define SYSTEM_FILE_NAME_MAX	40
+#define MAX_BLOCKS		50
 
 typedef void (*PrintFunc) (void *buf);
 typedef gboolean (*WriteFunc) (char **data, void *buf);
@@ -61,12 +62,16 @@ static void do_extent (char **args);
 static void do_chroot (char **args);
 static void do_slotmap (char **args);
 static void do_encode_lockres (char **args);
+static void do_decode_lockres (char **args);
 static void do_locate (char **args);
 static void do_fs_locks (char **args);
+static void do_bmap (char **args);
+static void do_icheck (char **args);
 
 dbgfs_gbls gbls;
 
 static Command commands[] = {
+	{ "bmap",	do_bmap },
 	{ "cat",	do_cat },
 	{ "cd",		do_cd },
 	{ "chroot",	do_chroot },
@@ -79,8 +84,11 @@ static Command commands[] = {
 	{ "help",	do_help },
 	{ "hb",		do_hb },
 	{ "?",		do_help },
+	{ "icheck",	do_icheck },
 	{ "lcd",	do_lcd },
 	{ "locate",	do_locate },
+	{ "ncheck",	do_locate },
+	{ "findpath",	do_locate },
 	{ "logdump",	do_logdump },
 	{ "ls",		do_ls },
 	{ "open",	do_open },
@@ -90,7 +98,8 @@ static Command commands[] = {
 	{ "slotmap",	do_slotmap },
 	{ "stat",	do_stat },
 	{ "stats",	do_stats },
-	{ "encode",	do_encode_lockres }
+	{ "encode",	do_encode_lockres },
+	{ "decode",	do_decode_lockres }
 };
 
 /*
@@ -150,9 +159,10 @@ void do_command (char *cmd)
 
 	fflush(stdout);
 
-	if (command)
+	if (command) {
+		gbls.cmd = command->cmd;
 		command->func (args);
-	else
+	} else
 		fprintf(stderr, "%s: command not found\n", args[0]);
 
 bail:
@@ -198,8 +208,7 @@ static int process_inode_args(char **args, uint64_t *blkno)
 	}
 
 	if (*blkno >= gbls.max_blocks) {
-		fprintf(stderr, "%s: Block number is larger than volume size\n",
-			args[0]);
+		com_err(args[0], OCFS2_ET_BAD_BLKNO, "- %"PRIu64"", *blkno);
 		return -1;
 	}
 
@@ -240,8 +249,7 @@ static int process_ls_args(char **args, uint64_t *blkno, int *long_opt)
 	}
 
 	if (*blkno >= gbls.max_blocks) {
-		fprintf(stderr, "%s: Block number is larger than volume size\n",
-			args[0]);
+		com_err(args[0], OCFS2_ET_BAD_BLKNO, "- %"PRIu64"", *blkno);
 		return -1;
 	}
 
@@ -250,25 +258,133 @@ static int process_ls_args(char **args, uint64_t *blkno, int *long_opt)
 
 /*
  * process_inodestr_args()
+ * 	args:	arguments starting from command
+ * 	count:	max space available in blkno
+ * 	blkno:	block nums extracted
+ * 	RETURN:	number of blknos
  *
  */
-static int process_inodestr_args(char **args, uint64_t *blkno)
+static int process_inodestr_args(char **args, int count, uint64_t *blkno)
 {
+	uint64_t *p;
+	int i;
+
 	if (check_device_open())
 		return -1;
 
-	if (!args[1] || inodestr_to_inode(args[1], blkno)) {
+	if (count < 1)
+		return 0;
+
+	for (i = 1, p = blkno; i < count + 1; ++i, ++p) {
+		if (!args[i] || inodestr_to_inode(args[i], p))
+			break;
+		if (*p >= gbls.max_blocks) {
+			com_err(args[0], OCFS2_ET_BAD_BLKNO, "- %"PRIu64"", *p);
+			return -1;
+		}
+	}
+
+	--i;
+
+	if (!i) {
 		fprintf(stderr, "usage: %s <inode#>\n", args[0]);
 		return -1;
 	}
 
-	if (*blkno >= gbls.max_blocks) {
-		fprintf(stderr, "%s: Block number is larger than volume size\n",
-			args[0]);
+	return  i;
+}
+
+/* open the device, read the block from the device and get the
+ * blocksize from the offset of the ocfs2_super_block.
+ */
+static errcode_t get_blocksize(char* dev, uint64_t offset, uint64_t *blocksize)
+{
+	errcode_t ret;
+	uint64_t blkno;
+	uint32_t blocksize_bits;
+	char *buf = NULL;
+	io_channel* channel = NULL;
+	struct ocfs2_dinode *di = NULL;
+
+	ret = io_open(dev, OCFS2_FLAG_RO, &channel);
+	if (ret)
+		goto bail;
+
+	/* since ocfs2_super_block inode can be stored in OCFS2_MIN_BLOCKSIZE,
+	 * so here we just use the minimum block size and read the information
+	 * in the specific offset.
+	 */
+	ret = io_set_blksize(channel, OCFS2_MIN_BLOCKSIZE);
+	if (ret)
+		goto bail;
+
+	ret = ocfs2_malloc_block(channel, &buf);
+	if (ret)
+		goto bail;
+
+	blkno = offset / OCFS2_MIN_BLOCKSIZE;
+	ret = io_read_block(channel, blkno, 1, buf);
+	if (ret)
+		goto bail;
+
+	di = (struct ocfs2_dinode *)buf;
+	blocksize_bits = le32_to_cpu(di->id2.i_super.s_blocksize_bits);
+	*blocksize = 1ULL << blocksize_bits;
+bail:
+	if (buf)
+		ocfs2_free(&buf);
+	if (channel)
+		io_close(channel);
+	return ret;
+}
+/*
+ * process_open_args
+ *
+ */
+static int process_open_args(char **args,
+			     uint64_t *superblock, uint64_t *blocksize)
+{
+	errcode_t ret = 0;
+	uint32_t s;
+	char *ptr;
+	uint64_t byte_off[OCFS2_MAX_BACKUP_SUPERBLOCKS], blksize;
+	int num, ind = 2;
+
+	if (!args[ind])
+		return 0;
+
+	if (args[ind] && !strcmp(args[ind], "-s"))
+		ind++;
+	else
+		return -1;
+
+	if(!args[ind])
+		return -1;
+
+	num = ocfs2_get_backup_super_offset(NULL,
+					    byte_off, ARRAY_SIZE(byte_off));
+	if (!num)
+		return -1;
+
+	s = strtoul(args[ind], &ptr, 0);
+	if (s < 1 || s > num) {
+		fprintf (stderr, "Backup super block is outside of valid range"
+			 "(between 1 and %d)\n", num);
 		return -1;
 	}
 
-	return 0;
+	ret = get_blocksize(args[1], byte_off[s-1], &blksize);
+	if (ret) {
+		com_err(args[0],ret, "Can't get the blocksize from the device"
+			" by the num %u\n", s);
+		goto bail;
+	}
+
+	*blocksize = blksize;
+	*superblock = byte_off[s-1]/blksize;
+	ret = 0;
+bail:
+	return ret;
 }
 
 /*
@@ -297,6 +413,64 @@ static int get_slotnum(char **args, uint16_t *slotnum)
 	return -1;
 }
 
+/*
+ * find_block_offset()
+ *
+ */
+static errcode_t find_block_offset(ocfs2_filesys *fs,
+				   struct ocfs2_extent_list *el,
+				   uint64_t blkoff, FILE *out)
+{
+	struct ocfs2_extent_block *eb;
+	struct ocfs2_extent_rec *rec;
+	errcode_t ret = 0;
+	char *buf = NULL;
+	int i;
+	uint32_t clstoff;
+	uint32_t tmp;
+
+	clstoff = ocfs2_blocks_to_clusters(fs, blkoff);
+
+	for (i = 0; i < el->l_next_free_rec; ++i) {
+		rec = &(el->l_recs[i]);
+
+		/* TODO Fix to handle sparse trees */
+		if (clstoff >= (rec->e_cpos + rec->e_clusters))
+			continue;
+
+		if (!el->l_tree_depth) {
+			tmp = blkoff - ocfs2_clusters_to_blocks(fs, clstoff);
+			dump_logical_blkno(out, rec->e_blkno + tmp);
+			goto bail;
+		}
+
+		ret = ocfs2_malloc_block(gbls.fs->fs_io, &buf);
+		if (ret) {
+			com_err(gbls.cmd, ret, "while allocating a block");
+			goto bail;
+		}
+
+		ret = ocfs2_read_extent_block(fs, rec->e_blkno, buf);
+		if (ret) {
+			com_err(gbls.cmd, ret, "while reading extent %"PRIu64,
+				rec->e_blkno);
+			goto bail;
+		}
+
+		eb = (struct ocfs2_extent_block *)buf;
+
+		ret = find_block_offset(fs, &(eb->h_list), blkoff, out);
+		goto bail;
+	}
+
+	dump_logical_blkno(out, 0);
+
+bail:
+	if (buf)
+		ocfs2_free(&buf);
+	return ret;
+}
+
 /*
  * traverse_extents()
  *
@@ -394,28 +568,30 @@ static void do_open (char **args)
 	char sysfile[SYSTEM_FILE_NAME_MAX];
 	int i;
 	struct ocfs2_super_block *sb;
+	uint64_t superblock = 0, block_size = 0;
 
 	if (gbls.device)
 		do_close (NULL);
 
-	if (dev == NULL) {
-		fprintf (stderr, "usage: %s <device>\n", args[0]);
+	if (dev == NULL || process_open_args(args, &superblock, &block_size)) {
+		fprintf (stderr, "usage: %s <device> [-s num]\n", args[0]);
 		return ;
 	}
 
 	flags = gbls.allow_write ? OCFS2_FLAG_RW : OCFS2_FLAG_RO;
         flags |= OCFS2_FLAG_HEARTBEAT_DEV_OK;
-	ret = ocfs2_open(dev, flags, 0, 0, &gbls.fs);
+	ret = ocfs2_open(dev, flags, superblock, block_size, &gbls.fs);
 	if (ret) {
 		gbls.fs = NULL;
-		com_err(args[0], ret, " ");
+		com_err(args[0], ret, "while opening context for device %s",
+			dev);
 		return ;
 	}
 
 	/* allocate blocksize buffer */
 	ret = ocfs2_malloc_block(gbls.fs->fs_io, &gbls.blockbuf);
 	if (ret) {
-		com_err(args[0], ret, " ");
+		com_err(args[0], ret, "while allocating a block");
 		return ;
 	}
 
@@ -475,7 +651,7 @@ static void do_close (char **args)
 
 	ret = ocfs2_close(gbls.fs);
 	if (ret)
-		com_err(args[0], ret, " ");
+		com_err(args[0], ret, "while closing context");
 	gbls.fs = NULL;
 
 	if (gbls.blockbuf)
@@ -501,7 +677,8 @@ static void do_cd (char **args)
 
 	ret = ocfs2_check_directory(gbls.fs, blkno);
 	if (ret) {
-		com_err(args[0], ret, " ");
+		com_err(args[0], ret, "while checking directory at "
+			"block %"PRIu64"", blkno);
 		return ;
 	}
 
@@ -524,7 +701,8 @@ static void do_chroot (char **args)
 
 	ret = ocfs2_check_directory(gbls.fs, blkno);
 	if (ret) {
-		com_err(args[0], ret, " ");
+		com_err(args[0], ret, "while checking directory at "
+			"blkno %"PRIu64"", blkno);
 		return ;
 	}
 
@@ -548,14 +726,15 @@ static void do_ls (char **args)
 
 	ret = ocfs2_check_directory(gbls.fs, blkno);
 	if (ret) {
-		com_err(args[0], ret, " ");
+		com_err(args[0], ret, "while checking directory at "
+			"block %"PRIu64"", blkno);
 		return ;
 	}
 
 	if (ls_opts.long_opt) {
 		ret = ocfs2_malloc_block(gbls.fs->fs_io, &ls_opts.buf);
 		if (ret) {
-			com_err(args[0], ret, " ");
+			com_err(args[0], ret, "while allocating a block");
 			return ;
 		}
 	}
@@ -564,7 +743,8 @@ static void do_ls (char **args)
 	ret = ocfs2_dir_iterate(gbls.fs, blkno, 0, NULL,
 				dump_dir_entry, (void *)&ls_opts);
 	if (ret)
-		com_err(args[0], ret, " ");
+		com_err(args[0], ret, "while iterating directory at "
+			"block %"PRIu64"", blkno);
 
 	close_pager(ls_opts.out);
 
@@ -580,22 +760,28 @@ static void do_ls (char **args)
  */
 static void do_help (char **args)
 {
+	printf ("bmap <filespec> <logical_blk>\t\tPrint the corresponding physical block# for the inode\n");
 	printf ("cat <filespec>\t\t\t\tPrints file on stdout\n");
 	printf ("cd <filespec>\t\t\t\tChange directory\n");
 	printf ("chroot <filespec>\t\t\tChange root\n");
 	printf ("close\t\t\t\t\tClose a device\n");
 	printf ("curdev\t\t\t\t\tShow current device\n");
+	printf ("decode <lockname#> ...\t\t\tDecode block#(s) from the lockname(s)\n");
 	printf ("dump [-p] <filespec> <outfile>\t\tDumps file to outfile on a mounted fs\n");
 	printf ("encode <filespec>\t\t\tShow lock name\n");
 	printf ("extent <block#>\t\t\t\tShow extent block\n");
+	printf ("findpath <block#>\t\t\tList one pathname of the inode/lockname\n");
 	printf ("fs_locks [-l]\t\t\t\tShow live fs locking state\n");
 	printf ("group <block#>\t\t\t\tShow chain group\n");
+	printf ("hb\t\t\t\t\tShows the used heartbeat blocks\n");
 	printf ("help, ?\t\t\t\t\tThis information\n");
+	printf ("icheck block# ...\t\t\tDo block->inode translation\n");
 	printf ("lcd <directory>\t\t\t\tChange directory on a mounted flesystem\n");
-	printf ("locate <filespec>\t\t\tList all paths to the inode\n");
+	printf ("locate <block#> ...\t\t\tList all pathnames of the inode(s)/lockname(s)\n");
 	printf ("logdump <slot#>\t\t\t\tPrints journal file for the node slot\n");
 	printf ("ls [-l] <filespec>\t\t\tList directory\n");
-	printf ("open <device>\t\t\t\tOpen a device\n");
+	printf ("ncheck <block#> ...\t\t\tList all pathnames of the inode(s)/lockname(s)\n");
+	printf ("open <device> [-s backup#]\t\t\t\tOpen a device\n");
 	printf ("quit, q\t\t\t\t\tExit the program\n");
 	printf ("rdump [-v] <filespec> <outdir>\t\tRecursively dumps from src to a dir on a mounted filesystem\n");
 	printf ("slotmap\t\t\t\t\tShow slot map\n");
@@ -628,7 +814,8 @@ static void do_lcd (char **args)
 	if (!args[1]) {
 		/* show cwd */
 		if (!getcwd(buf, sizeof(buf))) {
-			com_err(args[0], errno, " ");
+			com_err(args[0], errno, "while reading current "
+				"working directory");
 			return ;
 		}
 		fprintf(stdout, "%s\n", buf);
@@ -658,23 +845,59 @@ static void do_curdev (char **args)
  */
 static void do_stats (char **args)
 {
-	char *opts = args[1];
 	FILE *out;
+	errcode_t ret;
 	struct ocfs2_dinode *in;
 	struct ocfs2_super_block *sb;
+	int c, argc;
+	int sb_num = 0;
+	int only_super = 0;
+	char *ptr = NULL;
+	char *stats_usage = "usage: stats [-h] [-s backup#]";
+	char *buf = gbls.blockbuf;
 
 	if (check_device_open())
 		goto bail;
 
+	for (argc = 0; (args[argc]); ++argc);
+	optind = 0;
+
+	while ((c = getopt(argc, args, "hs:")) != -1) {
+		switch (c) {
+		case 'h':
+			only_super = 1;
+			break;
+		case 's':
+			sb_num = strtoul(optarg, &ptr, 0);
+			if (!ptr || *ptr) {
+				fprintf(stderr, "%s\n", stats_usage);
+				goto bail;
+			}
+			break;
+		default:
+			break;
+		}
+	}
+
+	if (!sb_num)
+		in = gbls.fs->fs_super;
+	else {
+		ret = ocfs2_read_backup_super(gbls.fs, sb_num, buf);
+		if (ret) {
+			com_err(gbls.cmd, ret, "while reading backup "
+				"superblock");
+			goto bail;
+		}
+		in = (struct ocfs2_dinode *)buf;
+	}
+
+	sb = OCFS2_RAW_SB(in);
+
 	out = open_pager(gbls.interactive);
-	in = gbls.fs->fs_super;
-	sb = OCFS2_RAW_SB(gbls.fs->fs_super);
 	dump_super_block(out, sb);
-
-	if (!opts || strcmp(opts, "-h"))
+	if (!only_super)
 		dump_inode(out, in);
-
-	close_pager (out);
+	close_pager(out);
 
 bail:
 	return ;
@@ -698,7 +921,7 @@ static void do_stat (char **args)
 	buf = gbls.blockbuf;
 	ret = ocfs2_read_inode(gbls.fs, blkno, buf);
 	if (ret) {
-		com_err(args[0], ret, " ");
+		com_err(args[0], ret, "while reading inode %"PRIu64"", blkno);
 		return ;
 	}
 
@@ -719,7 +942,8 @@ static void do_stat (char **args)
 		ret = traverse_extents(gbls.fs, &(inode->id2.i_list), out);
 
 	if (ret)
-		com_err(args[0], ret, " ");
+		com_err(args[0], ret, "while traversing inode at block "
+			"%"PRIu64, blkno);
 
 	close_pager(out);
 
@@ -741,7 +965,7 @@ static void do_hb (char **args)
 
 	ret = ocfs2_read_whole_file(gbls.fs, gbls.hb_blkno, &hbbuf, &len);
 	if (ret) {
-		com_err(args[0], ret, " ");
+		com_err(args[0], ret, "while reading heartbeat system file");
 		goto bail;
 	}
 
@@ -796,7 +1020,7 @@ static void do_dump (char **args)
 	ret = string_to_inode(gbls.fs, gbls.root_blkno, gbls.cwd_blkno,
 			      in_fn, &blkno);
 	if (ret) {
-		com_err(args[0], ret, " ");
+		com_err(args[0], ret, "'%s'", in_fn);
 		return ;
 	}
 
@@ -808,7 +1032,7 @@ static void do_dump (char **args)
 
 	ret = dump_file(gbls.fs, blkno, fd, out_fn, preserve);
 	if (ret)
-		com_err(args[0], ret, " ");
+		com_err(args[0], ret, "while dumping file");
 
 	return;
 }
@@ -830,7 +1054,7 @@ static void do_cat (char **args)
 	buf = gbls.blockbuf;
 	ret = ocfs2_read_inode(gbls.fs, blkno, buf);
 	if (ret) {
-		com_err(args[0], ret, " ");
+		com_err(args[0], ret, "while reading inode %"PRIu64"", blkno);
 		return ;
 	}
 
@@ -842,23 +1066,18 @@ static void do_cat (char **args)
 
 	ret = dump_file(gbls.fs, blkno, fileno(stdout),  NULL, 0);
 	if (ret)
-		com_err(args[0], ret, " ");
+		com_err(args[0], ret, "while reading file for inode %"PRIu64"",
+			blkno);
 
 	return ;
 }
 
-/*
- * do_logdump()
- *
- */
 static void do_logdump (char **args)
 {
-	char *logbuf = NULL;
-	uint64_t blkno = 0;
-	int32_t len = 0;
-	FILE *out;
+	errcode_t ret;
 	uint16_t slotnum;
-	errcode_t ret = 0;
+	uint64_t blkno;
+	FILE *out;
 
 	if (check_device_open())
 		return ;
@@ -867,21 +1086,14 @@ static void do_logdump (char **args)
 		return ;
 
 	blkno = gbls.jrnl_blkno[slotnum];
-	ret = read_whole_file(gbls.fs, blkno, &logbuf, &len);
-	if (ret) {
-		com_err(args[0], ret, " ");
-		goto bail;
-	}
 
 	out = open_pager(gbls.interactive);
-	read_journal (out, logbuf, (uint64_t)len);
+	ret = read_journal(gbls.fs, blkno, out);
 	close_pager (out);
+	if (ret)
+		com_err(gbls.cmd, ret, "while reading journal");
 
-bail:
-	if (logbuf)
-		ocfs2_free(&logbuf);
-
-	return ;
+	return;
 }
 
 /*
@@ -897,7 +1109,7 @@ static void do_group (char **args)
 	errcode_t ret = 0;
 	int index = 0;
 
-	if (process_inodestr_args(args, &blkno))
+	if (process_inodestr_args(args, 1, &blkno) != 1)
 		return ;
 
 	buf = gbls.blockbuf;
@@ -905,7 +1117,8 @@ static void do_group (char **args)
 	while (blkno) {
 		ret = ocfs2_read_group_desc(gbls.fs, blkno, buf);
 		if (ret) {
-			com_err(args[0], ret, " ");
+			com_err(args[0], ret, "while reading block group "
+				"descriptor %"PRIu64"", blkno);
 			close_pager (out);
 			return ;
 		}
@@ -933,13 +1146,14 @@ static void do_extent (char **args)
 	FILE *out;
 	errcode_t ret = 0;
 
-	if (process_inodestr_args(args, &blkno))
+	if (process_inodestr_args(args, 1, &blkno) != 1)
 		return ;
 
 	buf = gbls.blockbuf;
 	ret = ocfs2_read_extent_block(gbls.fs, blkno, buf);
 	if (ret) {
-		com_err(args[0], ret, " ");
+		com_err(args[0], ret, "while reading extent block %"PRIu64"",
+			blkno);
 		return ;
 	}
 
@@ -970,7 +1184,7 @@ static void do_slotmap (char **args)
 	/* read in the first block of the slot_map file */
 	ret = read_whole_file(gbls.fs, gbls.slotmap_blkno, &buf, &len);
 	if (ret) {
-		com_err(args[0], ret, " ");
+		com_err(args[0], ret, "while reading slotmap system file");
 		goto bail;
 	}
 
@@ -1022,7 +1236,7 @@ static void do_rdump(char **args)
 	ret = string_to_inode(gbls.fs, gbls.root_blkno, gbls.cwd_blkno,
 			      args[ind], &blkno);
 	if (ret) {
-		com_err(args[0], ret, " ");
+		com_err(args[0], ret, "while translating %s", args[ind]);
 		return ;
 	}
 
@@ -1070,7 +1284,8 @@ static void do_rdump(char **args)
 
 	ret = rdump_inode(gbls.fs, blkno, p, args[ind+1], verbose);
 	if (ret)
-		com_err(args[0], ret, " ");
+		com_err(args[0], ret, "while recursively dumping "
+			"inode %"PRIu64, blkno);
 
 	return ;
 }
@@ -1111,7 +1326,7 @@ static void do_encode_lockres (char **args)
 	}
 
 	if (ret) {
-		com_err(args[0], ret, " ");
+		com_err(args[0], ret, "while reading inode %"PRIu64"", blkno);
 		return ;
 	}
 
@@ -1129,18 +1344,42 @@ static void do_encode_lockres (char **args)
 	return ;
 }
 
+/*
+ * do_decode_lockres()
+ *
+ */
+static void do_decode_lockres(char **args)
+{
+	uint64_t blkno[MAX_BLOCKS];
+	int count;
+	int i;
+
+	count = process_inodestr_args(args, MAX_BLOCKS, blkno);
+	if (count < 1)
+		return ;
+
+	for (i = 0; i < count; ++i)
+		printf("\t%s\t%"PRIu64"\n", args[i + 1], blkno[i]);
+}
+
 /*
  * do_locate()
  *
  */
 static void do_locate(char **args)
 {
-	uint64_t blkno;
-
-	if (process_inode_args(args, &blkno))
+	uint64_t blkno[MAX_BLOCKS];
+	int count = 0;
+	int findall = 1;
+	
+	count = process_inodestr_args(args, MAX_BLOCKS, blkno);
+	if (count < 1)
 		return ;
 
-	find_inode_paths(gbls.fs, args, blkno, stdout);
+	if (!strncasecmp(args[0], "findpath", 8))
+		findall = 0;
+
+	find_inode_paths(gbls.fs, args, findall, count, blkno, stdout);
 }
 
 /*
@@ -1162,3 +1401,120 @@ static void do_fs_locks(char **args)
 	dump_fs_locks(gbls.fs->uuid_str, out, dump_lvbs);
 	close_pager(out);
 }
+
+/*
+ * do_bmap()
+ *
+ */
+static void do_bmap(char **args)
+{
+	struct ocfs2_dinode *inode;
+	const char *bmap_usage = "usage: bmap <filespec> <logical_blk>";
+	uint64_t blkno;
+	uint64_t blkoff;
+	char *endptr;
+	char *buf = NULL;
+	FILE *out;
+	errcode_t ret = 1;
+ 
+	if (check_device_open())
+		return;
+
+	if (!args[1]) {
+		fprintf(stderr, "usage: %s\n", bmap_usage);
+		return;
+	}
+
+	if (!args[1] || !args[2]) {
+		fprintf(stderr, "usage: %s\n", bmap_usage);
+		return;
+	}
+
+	ret = string_to_inode(gbls.fs, gbls.root_blkno, gbls.cwd_blkno,
+			      args[1], &blkno);
+	if (ret) {
+		com_err(args[0], ret, "'%s'", args[1]);
+		return;
+	}
+
+	if (blkno >= gbls.max_blocks) {
+		com_err(args[0], OCFS2_ET_BAD_BLKNO, "- %"PRIu64"", blkno);
+		return;
+	}
+
+	blkoff = strtoull(args[2], &endptr, 0);
+	if (*endptr) {
+		com_err(args[0], OCFS2_ET_BAD_BLKNO, "- %"PRIu64"", blkoff);
+		return;
+	}
+ 
+	buf = gbls.blockbuf;
+	ret = ocfs2_read_inode(gbls.fs, blkno, buf);
+	if (ret) {
+		com_err(args[0], ret, "while reading inode %"PRIu64, blkno);
+		return;
+	}
+ 
+	inode = (struct ocfs2_dinode *)buf;
+ 
+	out = open_pager(gbls.interactive);
+ 
+	if (inode->i_flags & (OCFS2_LOCAL_ALLOC_FL | OCFS2_CHAIN_FL |
+			      OCFS2_DEALLOC_FL)) {
+		dump_logical_blkno(out, 0);
+		goto bail;
+	}
+
+	if (S_ISLNK(inode->i_mode) && !inode->i_clusters) {
+		dump_logical_blkno(out, 0);
+		goto bail;
+	}
+
+	if (blkoff > (inode->i_size >>
+		      OCFS2_RAW_SB(gbls.fs->fs_super)->s_blocksize_bits)) {
+		dump_logical_blkno(out, 0);
+		goto bail;
+	}
+
+	find_block_offset(gbls.fs, &(inode->id2.i_list), blkoff, out);
+
+bail:
+	close_pager(out);
+ 
+	return;
+}
+
+static void do_icheck(char **args)
+{
+	const char *testb_usage = "usage: icheck block# ...";
+	char *endptr;
+	uint64_t blkno[MAX_BLOCKS];
+	int i;
+	FILE *out;
+
+	if (!args[1]) {
+		fprintf(stderr, "%s\n", testb_usage);
+		return;
+	}
+
+	for (i = 0; i < MAX_BLOCKS && args[i + 1]; ++i) {
+		blkno[i] = strtoull(args[i + 1], &endptr, 0);
+		if (*endptr) {
+			com_err(args[0], OCFS2_ET_BAD_BLKNO, "- %s", args[i + 1]);
+			return;
+		}
+
+		if (blkno[i] >= gbls.max_blocks) {
+			com_err(args[0], OCFS2_ET_BAD_BLKNO, "- %"PRIu64"", blkno[i]);
+			return;
+		}
+	}
+
+	out = open_pager(gbls.interactive);
+
+	find_block_inode(gbls.fs, blkno, i, out);
+
+	close_pager(out);
+
+	return;
+}
