@@ -70,6 +70,27 @@ struct chain_state {
 	uint16_t	cs_cpg;
 };
 
+static void find_max_free_bits(struct ocfs2_group_desc *gd, int *max_free_bits)
+{
+	int end = 0;
+	int start;
+	int free_bits;
+
+	*max_free_bits = 0;
+
+	while (end < gd->bg_bits) {
+		start = ocfs2_find_next_bit_clear(gd->bg_bitmap,
+						  gd->bg_bits, end);
+		if (start >= gd->bg_bits)
+			break;
+
+		end = ocfs2_find_next_bit_set(gd->bg_bitmap,
+					      gd->bg_bits, start);
+		free_bits = end - start;
+		*max_free_bits += free_bits;
+	}
+}
+
 static errcode_t repair_group_desc(o2fsck_state *ost,
 				   struct ocfs2_dinode *di,
 				   struct chain_state *cs,
@@ -78,6 +99,7 @@ static errcode_t repair_group_desc(o2fsck_state *ost,
 {
 	errcode_t ret = 0;
 	int changed = 0;
+	int max_free_bits = 0;
 
 	verbosef("checking desc at %"PRIu64"; blkno %"PRIu64" size %u bits %u "
 		 "free_bits %u chain %u generation %u\n", blkno, bg->bg_blkno,
@@ -128,13 +150,16 @@ static errcode_t repair_group_desc(o2fsck_state *ost,
 		changed = 1;
 	}
 
-	if ((bg->bg_free_bits_count > bg->bg_bits) &&
+	find_max_free_bits(bg, &max_free_bits);
+
+	if ((bg->bg_free_bits_count > max_free_bits) &&
 	    prompt(ost, PY, PR_GROUP_FREE_BITS,
 		   "Group descriptor at block %"PRIu64" claims to "
-		   "have %u free bits which is more than its %u total bits. "
+		   "have %u free bits which is more than %u bits"
+		   " indicated by the bitmap. "
 		   "Drop its free bit count down to the total?", blkno,
-		   bg->bg_free_bits_count, bg->bg_bits)) {
-		bg->bg_free_bits_count = bg->bg_bits;
+		   bg->bg_free_bits_count, max_free_bits)) {
+		bg->bg_free_bits_count = max_free_bits;
 		changed = 1;
 	}
 
@@ -233,6 +258,8 @@ static void unlink_group_desc(o2fsck_state *ost,
 			unlink = 1;
 			break;
 		}
+		if (unlink)
+			break;
 	}
 
 	if (!unlink)
@@ -278,6 +305,66 @@ static void mark_group_used(o2fsck_state *ost, struct chain_state *cs,
 				clusters);
 }
 
+/*
+ * Due to a glitch in mkfs, cl->cl_cpg for the GLOBAL BITMAP could be
+ * less than the max possible for volumes having just one cluster
+ * group. Fix.
+ */
+static errcode_t maybe_fix_clusters_per_group(o2fsck_state *ost,
+					      struct ocfs2_dinode *di)
+{
+	struct ocfs2_chain_list *cl;
+	struct ocfs2_group_desc *gd;
+	uint16_t new_cl_cpg = 0;
+	uint64_t blkno;
+	char *buf = NULL;
+	int ret = 0;
+
+	cl = &(di->id2.i_chain);
+	if (cl->cl_next_free_rec > 1)
+		goto out;
+
+	ret = ocfs2_malloc_block(ost->ost_fs->fs_io, &buf);
+	if (ret) {
+		com_err(whoami, ret, "while allocating block buffers "
+			"to fix cl_cpg");
+		goto out;
+	}
+	gd = (struct ocfs2_group_desc *) buf;
+
+	blkno = cl->cl_recs[0].c_blkno;
+
+	ret = ocfs2_read_group_desc(ost->ost_fs, blkno, (char *)gd);
+	if (ret) {
+		com_err(whoami, ret, "while reading group descriptor "
+			"at block %"PRIu64" to fix cl_cpg", blkno);
+		goto out;
+	}
+
+	new_cl_cpg = 8 * gd->bg_size;
+	if (cl->cl_cpg == new_cl_cpg)
+		goto out;
+
+	if (prompt(ost, PY, PR_CHAIN_CPG,
+		   "Global bitmap at block %"PRIu64" has clusters per group "
+		   "set to %u instead of %u. Fix?", di->i_blkno, cl->cl_cpg,
+		   new_cl_cpg)) {
+		cl->cl_cpg = new_cl_cpg;
+		ret = ocfs2_write_inode(ost->ost_fs, di->i_blkno, (char *)di);
+		if (ret) {
+			com_err(whoami, ret, "while writing inode alloc inode "
+				"%"PRIu64" to fix cl_cpg", di->i_blkno);
+			ost->ost_saw_error = 1;
+			ret = 0;
+		}
+	}
+
+out:
+	if (buf)
+		ocfs2_free(&buf);
+	return ret;
+}
+
 /* this takes a slightly ridiculous number of arguments :/ */
 static errcode_t check_chain(o2fsck_state *ost,
 			     struct ocfs2_dinode *di,
@@ -293,7 +380,7 @@ static errcode_t check_chain(o2fsck_state *ost,
 	struct ocfs2_group_desc *bg2 = (struct ocfs2_group_desc *)buf2;
 	uint64_t blkno;
 	errcode_t ret = 0;
-	int changed = 0, depth = 0, clear_ref = 0;
+	int depth = 0, clear_ref = 0;
 
 	verbosef("free %u total %u blkno %"PRIu64"\n", chain->c_free,
 		 chain->c_total, chain->c_blkno);
@@ -401,10 +488,10 @@ static errcode_t check_chain(o2fsck_state *ost,
 	 * ref we were working from */
 	if (clear_ref) {
 		if (depth == 0) {
-			chain->c_blkno = blkno;
+			chain->c_blkno = 0;
 			*chain_changed = 1;
 		} else {
-			bg1->bg_next_group = blkno;
+			bg1->bg_next_group = 0;
 			ret = ocfs2_write_group_desc(ost->ost_fs,
 					             bg1->bg_blkno,
 						     (char *)bg1);
@@ -432,7 +519,7 @@ static errcode_t check_chain(o2fsck_state *ost,
 			   cs->cs_total_bits)) {
 			chain->c_total = cs->cs_total_bits;
 			chain->c_free = cs->cs_free_bits;
-			changed = 1;
+			*chain_changed = 1;
 		}
 	}
 
@@ -576,6 +663,8 @@ static errcode_t verify_chain_alloc(o2fsck_state *ost,
 			 * update */
 			if (i < (cl->cl_next_free_rec - 1)) {
 				*cr = cl->cl_recs[cl->cl_next_free_rec - 1];
+				memset(&cl->cl_recs[cl->cl_next_free_rec - 1],
+					0, sizeof(struct ocfs2_chain_rec));
 				i--;
 			}
 
@@ -587,6 +676,9 @@ static errcode_t verify_chain_alloc(o2fsck_state *ost,
 
 	}
 
+	for (i = cl->cl_next_free_rec; i < cl->cl_count; i++)
+		memset(&cl->cl_recs[i], 0, sizeof(struct ocfs2_chain_rec));
+
 	if (di->id1.bitmap1.i_total != total || 
 	    (di->id1.bitmap1.i_used != total - free)) {
 		if (prompt(ost, PY, PR_CHAIN_GROUP_BITS,
@@ -883,12 +975,23 @@ errcode_t o2fsck_pass0(o2fsck_state *ost)
 	verbosef("found inode alloc %"PRIu64" at block %"PRIu64"\n",
 		 di->i_blkno, blkno);
 
+	ret = maybe_fix_clusters_per_group(ost, di);
+	if (ret)
+		goto out;
+
 	ret = verify_bitmap_descs(ost, di, blocks + ost->ost_fs->fs_blocksize,
 				  blocks + (ost->ost_fs->fs_blocksize * 2));
 
 	if (ret)
 		goto out;
 
+	if (fs->fs_super->i_clusters != di->i_clusters) {
+		if (prompt(ost, PY, PR_SUPERBLOCK_CLUSTERS,
+			   "Superblock has clusters set to %u instead of %u. Fix?",
+			   fs->fs_super->i_clusters, di->i_clusters))
+			ost->ost_num_clusters = di->i_clusters;
+	}
+
 	printf("Pass 0b: Checking inode allocation chains\n");
 
 	/* first the global inode alloc and then each of the node's
