@@ -570,7 +570,7 @@ static void vb_saw_block(struct verifying_blocks *vb, uint64_t bcount)
 static errcode_t process_link_block(struct verifying_blocks *vb,
 				    uint64_t blkno)
 {
-	char *buf, *null;
+	char *buf = NULL, *null;
 	errcode_t ret = 0;
 	unsigned int blocksize = vb->vb_ost->ost_fs->fs_blocksize;
 
@@ -597,7 +597,8 @@ static errcode_t process_link_block(struct verifying_blocks *vb,
 	}
 
 out:
-	ocfs2_free(&buf);
+	if (buf)
+		ocfs2_free(&buf);
 	return ret;
 }
 
@@ -712,6 +713,25 @@ static int verify_block(ocfs2_filesys *fs,
 	return 0;
 }
 
+/* This may be accessed many times for the same cluster.
+ * currently I haven't find a good way to avoid this since
+ * we use ocfs2_block_iterate_inode to clear the clusters.
+ * Anyway, it works well in clearing clusters.
+ */
+static int clear_block(ocfs2_filesys *fs,
+			uint64_t blkno,
+			uint64_t bcount,
+			void *priv_data)
+{
+	struct verifying_blocks *vb = priv_data;
+	o2fsck_state *ost = vb->vb_ost;
+	uint32_t clusters = ocfs2_blocks_to_clusters(ost->ost_fs, blkno);
+	
+	o2fsck_mark_cluster_unallocated(ost, clusters);
+
+	return 0;
+}
+
 /*
  * this verifies i_size and i_clusters for inodes that use i_list to
  * reference extents of data.
@@ -761,18 +781,27 @@ static errcode_t o2fsck_check_blocks(ocfs2_filesys *fs, o2fsck_state *ost,
 	}
 
 	/*
-	 * XXX we should have a helper that clears an inode and backs it out of
-	 * any book-keeping that it might have been included in, as though it
-	 * was never seen.  the alternative is to restart pass1 which seems
-	 * goofy. 
+	 * We have a helper function,clear_block, that clears an inode and 
+	 * backs it out of any book-keeping that it might have been included
+	 * in, as though it was never seen.
 	 */
 	if (vb.vb_clear) {
 		di->i_links_count = 0;
 		o2fsck_icount_set(ost->ost_icount_in_inodes, di->i_blkno,
 				  di->i_links_count);
 		di->i_dtime = time(NULL);
+		/* clear valid flag and stuff. */
+		ret = ocfs2_block_iterate_inode(fs, di,
+					        OCFS2_BLOCK_FLAG_APPEND,
+						clear_block, &vb);
+		di->i_flags &= !OCFS2_VALID_FL;
 		o2fsck_write_inode(ost, di->i_blkno, di);
-		/* XXX clear valid flag and stuff? */
+		/* for a directory, we also need to clear it 
+		 * from the dir_parent rb-tree. */
+		if (S_ISDIR(di->i_mode))
+			ocfsck_remove_dir_parent(&ost->ost_dir_parents,
+						 di->i_blkno);
+		goto out;	
 	}
 
 	if (vb.vb_num_blocks > 0)
@@ -994,6 +1023,18 @@ static errcode_t force_cluster_bit(o2fsck_state *ost,
 	return ret;
 }
 
+static inline int bit_in_backup_super(uint64_t bit,
+				      uint32_t *clusters, int len)
+{
+	int i;
+
+	for (i = 0; i < len; i++) {
+		if (bit == clusters[i])
+			return 1;
+	}
+	return 0;
+}
+
 /* once we've iterated all the inodes we should have the current working
  * set of which blocks we think are in use.  we use this to derive the set
  * of clusters that should be allocated in the cluster chain allocators.  we
@@ -1004,6 +1045,9 @@ static void write_cluster_alloc(o2fsck_state *ost)
 	errcode_t ret;
 	uint64_t blkno, last_cbit, cbit, cbit_found;
 	struct ocfs2_cluster_group_sizes cgs;
+	uint64_t blocks[OCFS2_MAX_BACKUP_SUPERBLOCKS];
+	uint32_t clusters[OCFS2_MAX_BACKUP_SUPERBLOCKS];
+	int backup_super = 0, num = 0, i;
 
 	ocfs2_calc_cluster_groups(ost->ost_fs->fs_clusters,
 				  ost->ost_fs->fs_blocksize, &cgs);
@@ -1032,6 +1076,22 @@ static void write_cluster_alloc(o2fsck_state *ost)
 		goto out;
 	}
 
+	/* handle the condition of backup superblock. */
+	memset(&blocks, 0, sizeof(blocks));
+	memset(&clusters, 0, sizeof(clusters));
+	if (OCFS2_HAS_COMPAT_FEATURE(OCFS2_RAW_SB(ost->ost_fs->fs_super),
+				     OCFS2_FEATURE_COMPAT_BACKUP_SB)) {
+		num = ocfs2_get_backup_super_offset(ost->ost_fs, blocks,
+					   ARRAY_SIZE(blocks));
+		if (num) {
+			backup_super = 1;
+			for (i = 0; i < num; i++)
+				clusters[i] =
+					ocfs2_blocks_to_clusters(ost->ost_fs,
+								 blocks[i]);
+		}
+	}
+
 	/* we walk our found blocks bitmap to find clusters that we think
 	 * are in use.  each time we find a block in a cluster we skip ahead
 	 * to the first block of the next cluster when looking for the next.
@@ -1043,6 +1103,10 @@ static void write_cluster_alloc(o2fsck_state *ost)
 	 * we special case the number of clusters as the cluster offset which
 	 * indicates that the rest of the bits to the end of the bitmap
 	 * should be clear.
+	 *
+	 * we should take backup superblock as a special case since it doesn't
+	 * belong to any inode. So it shouldn't be exist in
+	 * ost->ost_allocated_clusters.
 	 */
 	for (last_cbit = 0, cbit = 0;
 	     cbit < ost->ost_fs->fs_clusters; 
@@ -1068,9 +1132,15 @@ static void write_cluster_alloc(o2fsck_state *ost)
 
 		/* clear set bits that should have been clear up to cbit */
 		while (cbit_found < cbit) {
-			force_cluster_bit(ost, ci, cbit_found, 0);
+			/* check whether the volume has backup blocks
+			 * and if yes, check whether the cluster contains
+			 * one of the backup blocks.
+			 */
+			if (!backup_super ||
+			    !bit_in_backup_super(cbit_found, clusters, num))
+				force_cluster_bit(ost, ci, cbit_found, 0);
 			cbit_found++;
-			ret = ocfs2_bitmap_find_next_set(ci->ci_chains, cbit, 
+			ret = ocfs2_bitmap_find_next_set(ci->ci_chains, cbit_found, 
 							 &cbit_found);
 			if (ret == OCFS2_ET_BIT_NOT_FOUND)
 				cbit_found = ost->ost_fs->fs_clusters;
