@@ -182,6 +182,7 @@ static void ocfs2_init_eb(ocfs2_filesys *fs,
 			  uint64_t gd_blkno, uint64_t blkno)
 {
 	strcpy(eb->h_signature, OCFS2_EXTENT_BLOCK_SIGNATURE);
+	eb->h_fs_generation = fs->fs_super->i_fs_generation;
 	eb->h_blkno = blkno;
 	eb->h_suballoc_slot = 0;
 	eb->h_suballoc_bit = (uint16_t)(blkno - gd_blkno);
@@ -214,7 +215,8 @@ errcode_t ocfs2_new_inode(ocfs2_filesys *fs, uint64_t *ino, int mode)
 						&gd_blkno, ino);
 		if (ret)
 			goto out;
-	}
+	} else if (ret)
+		goto out;
 
 	memset(buf, 0, fs->fs_blocksize);
 	di = (struct ocfs2_dinode *)buf;
@@ -360,7 +362,15 @@ errcode_t ocfs2_new_extent_block(ocfs2_filesys *fs, uint64_t *blkno)
 
 	ret = ocfs2_chain_alloc_with_io(fs, fs->fs_eb_allocs[0],
 					&gd_blkno, blkno);
-	if (ret)
+	if (ret == OCFS2_ET_BIT_NOT_FOUND) {
+		ret = ocfs2_chain_add_group(fs, fs->fs_eb_allocs[0]);
+		if (ret)
+			goto out;
+		ret = ocfs2_chain_alloc_with_io(fs, fs->fs_eb_allocs[0],
+						&gd_blkno, blkno);
+		if (ret)
+			goto out;
+	} else if (ret)
 		goto out;
 
 	memset(buf, 0, fs->fs_blocksize);
