@@ -39,6 +39,7 @@
 #include <sys/stat.h>
 #include <time.h>
 #include <string.h>
+#include <stddef.h>
 
 #include <limits.h>
 
@@ -48,17 +49,23 @@
 
 #include "byteorder.h"
 
-#include <kernel-list.h>
-#include <kernel-rbtree.h>
-
-#include <o2dlm.h>
-#include <o2cb.h>
-
 #if OCFS2_FLAT_INCLUDES
+#include "o2dlm.h"
+#include "o2cb.h"
+
+#include "kernel-list.h"
+#include "kernel-rbtree.h"
+
 #include "ocfs2_err.h"
 #include "ocfs2_fs.h"
 #include "jbd.h"
 #else
+#include <o2dlm/o2dlm.h>
+#include <o2cb/o2cb.h>
+
+#include <ocfs2/kernel-list.h>
+#include <ocfs2/kernel-rbtree.h>
+
 #include <ocfs2/ocfs2_err.h>
 #include <ocfs2/ocfs2_fs.h>
 #include <ocfs2/jbd.h>
@@ -66,9 +73,18 @@
 
 #include "ocfs2_lockid.h"
 
-#define OCFS2_LIB_FEATURE_INCOMPAT_SUPP		(OCFS2_FEATURE_INCOMPAT_SUPP | OCFS2_FEATURE_INCOMPAT_HEARTBEAT_DEV)
+#define OCFS2_LIB_FEATURE_INCOMPAT_SUPP		(OCFS2_FEATURE_INCOMPAT_SUPP | \
+						 OCFS2_FEATURE_INCOMPAT_HEARTBEAT_DEV | \
+						 OCFS2_FEATURE_INCOMPAT_RESIZE_INPROG | \
+						 OCFS2_FEATURE_INCOMPAT_LOCAL_MOUNT)
+
 #define OCFS2_LIB_FEATURE_RO_COMPAT_SUPP	OCFS2_FEATURE_RO_COMPAT_SUPP
 
+#define OCFS2_LIB_FEATURE_COMPAT_SUPP		OCFS2_FEATURE_COMPAT_SUPP
+
+/* define OCFS2_SB for ocfs2-tools */
+#define OCFS2_SB(sb)	(sb)
+
 /* Flags for the ocfs2_filesys structure */
 #define OCFS2_FLAG_RO			0x00
 #define OCFS2_FLAG_RW			0x01
@@ -80,6 +96,7 @@
 						   vol_header structure
 						   for revision info */
 #define OCFS2_FLAG_HEARTBEAT_DEV_OK	0x40
+#define OCFS2_FLAG_STRICT_COMPAT_CHECK	0x80
 
 /* Return flags for the extent iterator functions */
 #define OCFS2_EXTENT_CHANGED	0x01
@@ -248,7 +265,9 @@ errcode_t io_read_block(io_channel *channel, int64_t blkno, int count,
 errcode_t io_write_block(io_channel *channel, int64_t blkno, int count,
 			 const char *data);
 
+errcode_t ocfs2_read_super(ocfs2_filesys *fs, uint64_t superblock, char *sb);
 errcode_t ocfs2_write_super(ocfs2_filesys *fs);
+int ocfs2_mount_local(ocfs2_filesys *fs);
 errcode_t ocfs2_open(const char *name, int flags,
 		     unsigned int superblock, unsigned int blksize,
 		     ocfs2_filesys **ret_fs);
@@ -300,13 +319,12 @@ errcode_t ocfs2_load_extent_map(ocfs2_filesys *fs,
 void ocfs2_swap_journal_superblock(journal_superblock_t *jsb);
 errcode_t ocfs2_init_journal_superblock(ocfs2_filesys *fs, char *buf,
 					int buflen, uint32_t jrnl_size);
-errcode_t ocfs2_create_journal_superblock(ocfs2_filesys *fs,
-					  uint32_t size, int flags,
-					  char **ret_jsb);
 errcode_t ocfs2_read_journal_superblock(ocfs2_filesys *fs, uint64_t blkno,
 					char *jsb_buf);
 errcode_t ocfs2_write_journal_superblock(ocfs2_filesys *fs, uint64_t blkno,
 					 char *jsb_buf);
+errcode_t ocfs2_make_journal(ocfs2_filesys *fs, uint64_t blkno,
+			     uint32_t clusters);
 
 errcode_t ocfs2_read_extent_block(ocfs2_filesys *fs, uint64_t blkno,
        				  char *eb_buf);
@@ -430,8 +448,9 @@ errcode_t ocfs2_bitmap_find_next_clear(ocfs2_bitmap *bitmap,
 errcode_t ocfs2_bitmap_read(ocfs2_bitmap *bitmap);
 errcode_t ocfs2_bitmap_write(ocfs2_bitmap *bitmap);
 uint64_t ocfs2_bitmap_get_set_bits(ocfs2_bitmap *bitmap);
-errcode_t ocfs2_bitmap_alloc_range(ocfs2_bitmap *bitmap, uint64_t len, 
-				   uint64_t *first_bit);
+errcode_t ocfs2_bitmap_alloc_range(ocfs2_bitmap *bitmap, uint64_t min,
+				   uint64_t len, uint64_t *first_bit,
+				   uint64_t *bits_found);
 errcode_t ocfs2_bitmap_clear_range(ocfs2_bitmap *bitmap, uint64_t len, 
 				   uint64_t first_bit);
 
@@ -484,8 +503,10 @@ errcode_t ocfs2_chain_free(ocfs2_filesys *fs,
 			   uint64_t bitno);
 errcode_t ocfs2_chain_alloc_range(ocfs2_filesys *fs,
 				  ocfs2_cached_inode *cinode,
+				  uint64_t min,
 				  uint64_t requested,
-				  uint64_t *start_bit);
+				  uint64_t *start_bit,
+				  uint64_t *bits_found);
 errcode_t ocfs2_chain_free_range(ocfs2_filesys *fs,
 				 ocfs2_cached_inode *cinode,
 				 uint64_t len,
@@ -531,8 +552,10 @@ errcode_t ocfs2_extend_allocation(ocfs2_filesys *fs, uint64_t ino,
 				  uint32_t new_clusters);
 errcode_t ocfs2_truncate(ocfs2_filesys *fs, uint64_t ino, uint64_t new_i_size);
 errcode_t ocfs2_new_clusters(ocfs2_filesys *fs,
+			     uint32_t min,
 			     uint32_t requested,
-			     uint64_t *start_blkno);
+			     uint64_t *start_blkno,
+			     uint32_t *clusters_found);
 errcode_t ocfs2_free_clusters(ocfs2_filesys *fs,
 			      uint32_t len,
 			      uint64_t start_blkno);
@@ -591,6 +614,31 @@ errcode_t ocfs2_encode_lockres(enum ocfs2_lock_type type, uint64_t blkno,
 errcode_t ocfs2_decode_lockres(char *lockres, int len, enum ocfs2_lock_type *type,
 			       uint64_t *blkno, uint32_t *generation);
 
+/* write the superblock at the specific block. */
+errcode_t ocfs2_write_backup_super(ocfs2_filesys *fs, uint64_t blkno);
+
+/* Get the blkno according to the file system info.
+ * The unused ones, depending on the volume size, are zeroed.
+ * Return the length of the block array.
+ */
+int ocfs2_get_backup_super_offset(ocfs2_filesys *fs,
+				  uint64_t *blocks, size_t len);
+
+/* This function will get the superblock pointed to by fs and copy it to
+ * the blocks. But first it will ensure all the appropriate clusters are free.
+ * If not, it will error out with ENOSPC. If free, it will set bits for all
+ * the clusters, zero the clusters and write the backup sb.
+ * In case of updating, it will override the backup blocks with the newest
+ * superblock information.
+ */
+errcode_t ocfs2_set_backup_super(ocfs2_filesys *fs,
+				 uint64_t *blocks, size_t len);
+
+/* Refresh the backup superblock inoformation. */
+errcode_t ocfs2_refresh_backup_super(ocfs2_filesys *fs,
+				     uint64_t *blocks, size_t len);
+
+errcode_t ocfs2_read_backup_super(ocfs2_filesys *fs, int backup, char *sbbuf);
 
 /* 
  * ${foo}_to_${bar} is a floor function.  blocks_to_clusters will
@@ -644,6 +692,21 @@ static inline uint32_t ocfs2_clusters_in_blocks(ocfs2_filesys *fs,
 	return (uint32_t)(ret >> c_to_b_bits);
 }
 
+/* given a cluster offset, calculate which block group it belongs to
+ * and return that block offset. */
+static inline uint64_t ocfs2_which_cluster_group(ocfs2_filesys *fs,
+						 uint16_t cpg,
+						 uint32_t cluster)
+{
+	struct ocfs2_super_block *sb = OCFS2_RAW_SB(fs->fs_super);
+	uint32_t group_no;
+
+	group_no = cluster / cpg;
+	if (!group_no)
+		return sb->s_first_cluster_group;
+	return ocfs2_clusters_to_blocks(fs, group_no * cpg);
+}
+
 static inline int ocfs2_block_out_of_range(ocfs2_filesys *fs, uint64_t block)
 {
 	return (block < OCFS2_SUPER_BLOCK_BLKNO) || (block > fs->fs_blocks);
@@ -691,4 +754,6 @@ static inline void ocfs2_calc_cluster_groups(uint64_t clusters,
 	(void) (&_x == &_y);            \
 	_x > _y ? _x : _y; })
 
+/* lifted from the kernel. include/linux/kernel.h */
+#define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))
 #endif  /* _FILESYS_H */
