@@ -85,8 +85,8 @@
 #define OCFS2_CLEAR_INCOMPAT_FEATURE(sb,mask)			\
 	OCFS2_SB(sb)->s_feature_incompat &= ~(mask)
 
-#define OCFS2_FEATURE_COMPAT_SUPP	0
-#define OCFS2_FEATURE_INCOMPAT_SUPP	0
+#define OCFS2_FEATURE_COMPAT_SUPP	OCFS2_FEATURE_COMPAT_BACKUP_SB
+#define OCFS2_FEATURE_INCOMPAT_SUPP	OCFS2_FEATURE_INCOMPAT_LOCAL_MOUNT
 #define OCFS2_FEATURE_RO_COMPAT_SUPP	0
 
 /*
@@ -96,6 +96,32 @@
  */
 #define OCFS2_FEATURE_INCOMPAT_HEARTBEAT_DEV	0x0002
 
+/*
+ * tunefs sets this incompat flag before starting the resize and clears it
+ * at the end. This flag protects users from inadvertently mounting the fs
+ * after an aborted run without fsck-ing.
+ */
+#define OCFS2_FEATURE_INCOMPAT_RESIZE_INPROG    0x0004
+
+/* Used to denote a non-clustered volume */
+#define OCFS2_FEATURE_INCOMPAT_LOCAL_MOUNT	0x0008
+
+/* Support for sparse allocation in b-trees */
+#define OCFS2_FEATURE_INCOMPAT_SPARSE_ALLOC	0x0010
+
+/*
+ * backup superblock flag is used to indicate that this volume
+ * has backup superblocks.
+ */
+#define OCFS2_FEATURE_COMPAT_BACKUP_SB		0x0001
+
+/* The byte offset of the first backup block will be 1G.
+ * The following will be 4G, 16G, 64G, 256G and 1T.
+ */
+#define OCFS2_BACKUP_SB_START			1 << 30
+
+/* the max backup superblock nums */
+#define OCFS2_MAX_BACKUP_SUPERBLOCKS	6
 
 /*
  * Flags on ocfs2_dinode.i_flags
@@ -114,6 +140,26 @@
 #define OCFS2_CHAIN_FL		(0x00000400)	/* Chain allocator */
 #define OCFS2_DEALLOC_FL	(0x00000800)	/* Truncate log */
 
+/* Inode attributes, keep in sync with EXT2 */
+#define OCFS2_SECRM_FL		(0x00000001)	/* Secure deletion */
+#define OCFS2_UNRM_FL		(0x00000002)	/* Undelete */
+#define OCFS2_COMPR_FL		(0x00000004)	/* Compress file */
+#define OCFS2_SYNC_FL		(0x00000008)	/* Synchronous updates */
+#define OCFS2_IMMUTABLE_FL	(0x00000010)	/* Immutable file */
+#define OCFS2_APPEND_FL		(0x00000020)	/* writes to file may only append */
+#define OCFS2_NODUMP_FL		(0x00000040)	/* do not dump file */
+#define OCFS2_NOATIME_FL	(0x00000080)	/* do not update atime */
+#define OCFS2_DIRSYNC_FL	(0x00010000)	/* dirsync behaviour (directories only) */
+
+#define OCFS2_FL_VISIBLE	(0x000100FF)	/* User visible flags */
+#define OCFS2_FL_MODIFIABLE	(0x000100FF)	/* User modifiable flags */
+
+/*
+ * ioctl commands
+ */
+#define OCFS2_IOC_GETFLAGS	_IOR('f', 1, long)
+#define OCFS2_IOC_SETFLAGS	_IOW('f', 2, long)
+
 /*
  * Journal Flags (ocfs2_dinode.id1.journal1.i_flags)
  */
@@ -138,7 +184,6 @@
 
 /* Journal limits (in bytes) */
 #define OCFS2_MIN_JOURNAL_SIZE		(4 * 1024 * 1024)
-#define OCFS2_MAX_JOURNAL_SIZE		(500 * 1024 * 1024)
 
 struct ocfs2_system_inode_info {
 	char	*si_name;
@@ -185,7 +230,8 @@ static struct ocfs2_system_inode_info ocfs2_system_inodes[NUM_SYSTEM_INODES] = {
 };
 
 /* Parameter passed from mount.ocfs2 to module */
-#define OCFS2_HB_OK		"hbok"
+#define OCFS2_HB_NONE			"heartbeat=none"
+#define OCFS2_HB_LOCAL			"heartbeat=local"
 
 /*
  * OCFS2 directory file types.  Only the low 3 bits are used.  The
@@ -237,23 +283,23 @@ static unsigned char ocfs2_type_by_mode[S_IFMT >> S_SHIFT] = {
  * On disk extent record for OCFS2
  * It describes a range of clusters on disk.
  */
-typedef struct _ocfs2_extent_rec {
+struct ocfs2_extent_rec {
 /*00*/	__le32 e_cpos;		/* Offset into the file, in clusters */
 	__le32 e_clusters;	/* Clusters covered by this extent */
 	__le64 e_blkno;		/* Physical disk offset, in blocks */
 /*10*/
-} ocfs2_extent_rec;
+};
 
-typedef struct _ocfs2_chain_rec {
+struct ocfs2_chain_rec {
 	__le32 c_free;	/* Number of free bits in this chain. */
 	__le32 c_total;	/* Number of total bits in this chain */
 	__le64 c_blkno;	/* Physical disk offset (blocks) of 1st group */
-} ocfs2_chain_rec;
+};
 
-typedef struct _ocfs2_truncate_rec {
+struct ocfs2_truncate_rec {
 	__le32 t_start;		/* 1st cluster in this log */
 	__le32 t_clusters;	/* Number of total clusters covered */
-} ocfs2_truncate_rec;
+};
 
 /*
  * On disk extent list for OCFS2 (node in the tree).  Note that this
@@ -261,7 +307,7 @@ typedef struct _ocfs2_truncate_rec {
  * offsets are relative to ocfs2_dinode.id2.i_list or
  * ocfs2_extent_block.h_list, respectively.
  */
-typedef struct _ocfs2_extent_list {
+struct ocfs2_extent_list {
 /*00*/	__le16 l_tree_depth;		/* Extent tree depth from this
 					   point.  0 means data extents
 					   hang directly off this
@@ -271,39 +317,39 @@ typedef struct _ocfs2_extent_list {
 	__le16 l_reserved1;
 	__le64 l_reserved2;		/* Pad to
 					   sizeof(ocfs2_extent_rec) */
-/*10*/	ocfs2_extent_rec l_recs[0];	/* Extent records */
-} ocfs2_extent_list;
+/*10*/	struct ocfs2_extent_rec l_recs[0];	/* Extent records */
+};
 
 /*
  * On disk allocation chain list for OCFS2.  Note that this is
  * contained inside ocfs2_dinode, so the offsets are relative to
  * ocfs2_dinode.id2.i_chain.
  */
-typedef struct _ocfs2_chain_list {
+struct ocfs2_chain_list {
 /*00*/	__le16 cl_cpg;			/* Clusters per Block Group */
 	__le16 cl_bpc;			/* Bits per cluster */
-	__le16 cl_count;			/* Total chains in this list */
-	__le16 cl_next_free_rec;		/* Next unused chain slot */
+	__le16 cl_count;		/* Total chains in this list */
+	__le16 cl_next_free_rec;	/* Next unused chain slot */
 	__le64 cl_reserved1;
-/*10*/	ocfs2_chain_rec cl_recs[0];	/* Chain records */
-} ocfs2_chain_list;
+/*10*/	struct ocfs2_chain_rec cl_recs[0];	/* Chain records */
+};
 
 /*
  * On disk deallocation log for OCFS2.  Note that this is
  * contained inside ocfs2_dinode, so the offsets are relative to
  * ocfs2_dinode.id2.i_dealloc.
  */
-typedef struct _ocfs2_truncate_log {
-/*00*/	__le16 tl_count;			/* Total records in this log */
+struct ocfs2_truncate_log {
+/*00*/	__le16 tl_count;		/* Total records in this log */
 	__le16 tl_used;			/* Number of records in use */
 	__le32 tl_reserved1;
-/*08*/	ocfs2_truncate_rec tl_recs[0];	/* Truncate records */
-} ocfs2_truncate_log;
+/*08*/	struct ocfs2_truncate_rec tl_recs[0];	/* Truncate records */
+};
 
 /*
  * On disk extent block (indirect block) for OCFS2
  */
-typedef struct _ocfs2_extent_block
+struct ocfs2_extent_block
 {
 /*00*/	__u8 h_signature[8];		/* Signature for verification */
 	__le64 h_reserved1;
@@ -317,16 +363,16 @@ typedef struct _ocfs2_extent_block
 	__le64 h_next_leaf_blk;		/* Offset on disk, in blocks,
 					   of next leaf header pointing
 					   to data */
-/*30*/	ocfs2_extent_list h_list;	/* Extent record list */
+/*30*/	struct ocfs2_extent_list h_list;	/* Extent record list */
 /* Actual on-disk size is one block */
-} ocfs2_extent_block;
+};
 
 /*
  * On disk superblock for OCFS2
  * Note that it is contained inside an ocfs2_dinode, so all offsets
  * are relative to the start of ocfs2_dinode.id2.
  */
-typedef struct _ocfs2_super_block {
+struct ocfs2_super_block {
 /*00*/	__le16 s_major_rev_level;
 	__le16 s_minor_rev_level;
 	__le16 s_mnt_count;
@@ -354,26 +400,26 @@ typedef struct _ocfs2_super_block {
 /*50*/	__u8  s_label[OCFS2_MAX_VOL_LABEL_LEN];	/* Label for mounting, etc. */
 /*90*/	__u8  s_uuid[OCFS2_VOL_UUID_LEN];	/* 128-bit uuid */
 /*A0*/
-} ocfs2_super_block;
+};
 
 /*
  * Local allocation bitmap for OCFS2 slots
  * Note that it exists inside an ocfs2_dinode, so all offsets are
  * relative to the start of ocfs2_dinode.id2.
  */
-typedef struct _ocfs2_local_alloc
+struct ocfs2_local_alloc
 {
 /*00*/	__le32 la_bm_off;	/* Starting bit offset in main bitmap */
 	__le16 la_size;		/* Size of included bitmap, in bytes */
 	__le16 la_reserved1;
 	__le64 la_reserved2;
 /*10*/	__u8   la_bitmap[0];
-} ocfs2_local_alloc;
+};
 
 /*
  * On disk inode for OCFS2
  */
-typedef struct _ocfs2_dinode {
+struct ocfs2_dinode {
 /*00*/	__u8 i_signature[8];		/* Signature for validation */
 	__le32 i_generation;		/* Generation number */
 	__le16 i_suballoc_slot;		/* Slot suballocator this inode
@@ -399,7 +445,9 @@ typedef struct _ocfs2_dinode {
 	__le32 i_atime_nsec;
 	__le32 i_ctime_nsec;
 	__le32 i_mtime_nsec;
-/*70*/	__le64 i_reserved1[9];
+	__le32 i_attr;
+	__le32 i_reserved1;
+/*70*/	__le64 i_reserved2[8];
 /*B8*/	union {
 		__le64 i_pad1;		/* Generic way to refer to this
 					   64bit union */
@@ -419,15 +467,15 @@ typedef struct _ocfs2_dinode {
 		} journal1;
 	} id1;				/* Inode type dependant 1 */
 /*C0*/	union {
-		ocfs2_super_block  i_super;
-		ocfs2_local_alloc  i_lab;
-		ocfs2_chain_list   i_chain;
-		ocfs2_extent_list  i_list;
-		ocfs2_truncate_log i_dealloc;
-		__u8               i_symlink[0];
+		struct ocfs2_super_block	i_super;
+		struct ocfs2_local_alloc	i_lab;
+		struct ocfs2_chain_list		i_chain;
+		struct ocfs2_extent_list	i_list;
+		struct ocfs2_truncate_log	i_dealloc;
+		__u8               		i_symlink[0];
 	} id2;
 /* Actual on-disk size is one block */
-} ocfs2_dinode;
+};
 
 /*
  * On-disk directory entry structure for OCFS2
@@ -446,7 +494,7 @@ struct ocfs2_dir_entry {
 /*
  * On disk allocator group structure for OCFS2
  */
-typedef struct _ocfs2_group_desc
+struct ocfs2_group_desc
 {
 /*00*/	__u8    bg_signature[8];        /* Signature for validation */
 	__le16   bg_size;                /* Size of included bitmap in
@@ -464,13 +512,13 @@ typedef struct _ocfs2_group_desc
 	__le64   bg_blkno;               /* Offset on disk, in blocks */
 /*30*/	__le64   bg_reserved2[2];
 /*40*/	__u8    bg_bitmap[0];
-} ocfs2_group_desc;
+};
 
 #ifdef __KERNEL__
 static inline int ocfs2_fast_symlink_chars(struct super_block *sb)
 {
 	return  sb->s_blocksize -
-		 offsetof(struct _ocfs2_dinode, id2.i_symlink);
+		 offsetof(struct ocfs2_dinode, id2.i_symlink);
 }
 
 static inline int ocfs2_extent_recs_per_inode(struct super_block *sb)
@@ -478,9 +526,9 @@ static inline int ocfs2_extent_recs_per_inode(struct super_block *sb)
 	int size;
 
 	size = sb->s_blocksize -
-		offsetof(struct _ocfs2_dinode, id2.i_list.l_recs);
+		offsetof(struct ocfs2_dinode, id2.i_list.l_recs);
 
-	return size / sizeof(struct _ocfs2_extent_rec);
+	return size / sizeof(struct ocfs2_extent_rec);
 }
 
 static inline int ocfs2_chain_recs_per_inode(struct super_block *sb)
@@ -488,9 +536,9 @@ static inline int ocfs2_chain_recs_per_inode(struct super_block *sb)
 	int size;
 
 	size = sb->s_blocksize -
-		offsetof(struct _ocfs2_dinode, id2.i_chain.cl_recs);
+		offsetof(struct ocfs2_dinode, id2.i_chain.cl_recs);
 
-	return size / sizeof(struct _ocfs2_chain_rec);
+	return size / sizeof(struct ocfs2_chain_rec);
 }
 
 static inline u16 ocfs2_extent_recs_per_eb(struct super_block *sb)
@@ -498,9 +546,9 @@ static inline u16 ocfs2_extent_recs_per_eb(struct super_block *sb)
 	int size;
 
 	size = sb->s_blocksize -
-		offsetof(struct _ocfs2_extent_block, h_list.l_recs);
+		offsetof(struct ocfs2_extent_block, h_list.l_recs);
 
-	return size / sizeof(struct _ocfs2_extent_rec);
+	return size / sizeof(struct ocfs2_extent_rec);
 }
 
 static inline u16 ocfs2_local_alloc_size(struct super_block *sb)
@@ -508,7 +556,7 @@ static inline u16 ocfs2_local_alloc_size(struct super_block *sb)
 	u16 size;
 
 	size = sb->s_blocksize -
-		offsetof(struct _ocfs2_dinode, id2.i_lab.la_bitmap);
+		offsetof(struct ocfs2_dinode, id2.i_lab.la_bitmap);
 
 	return size;
 }
@@ -518,7 +566,7 @@ static inline int ocfs2_group_bitmap_size(struct super_block *sb)
 	int size;
 
 	size = sb->s_blocksize -
-		offsetof(struct _ocfs2_group_desc, bg_bitmap);
+		offsetof(struct ocfs2_group_desc, bg_bitmap);
 
 	return size;
 }
@@ -528,14 +576,28 @@ static inline int ocfs2_truncate_recs_per_inode(struct super_block *sb)
 	int size;
 
 	size = sb->s_blocksize -
-		offsetof(struct _ocfs2_dinode, id2.i_dealloc.tl_recs);
+		offsetof(struct ocfs2_dinode, id2.i_dealloc.tl_recs);
+
+	return size / sizeof(struct ocfs2_truncate_rec);
+}
+
+static inline u64 ocfs2_backup_super_blkno(struct super_block *sb, int index)
+{
+	u64 offset = OCFS2_BACKUP_SB_START;
+
+	if (index >= 0 && index < OCFS2_MAX_BACKUP_SUPERBLOCKS) {
+		offset <<= (2 * index);
+		offset /= sb->s_blocksize;
+		return offset;
+	}
+
+	return 0;
 
-	return size / sizeof(struct _ocfs2_truncate_rec);
 }
 #else
 static inline int ocfs2_fast_symlink_chars(int blocksize)
 {
-	return blocksize - offsetof(struct _ocfs2_dinode, id2.i_symlink);
+	return blocksize - offsetof(struct ocfs2_dinode, id2.i_symlink);
 }
 
 static inline int ocfs2_extent_recs_per_inode(int blocksize)
@@ -543,9 +605,9 @@ static inline int ocfs2_extent_recs_per_inode(int blocksize)
 	int size;
 
 	size = blocksize -
-		offsetof(struct _ocfs2_dinode, id2.i_list.l_recs);
+		offsetof(struct ocfs2_dinode, id2.i_list.l_recs);
 
-	return size / sizeof(struct _ocfs2_extent_rec);
+	return size / sizeof(struct ocfs2_extent_rec);
 }
 
 static inline int ocfs2_chain_recs_per_inode(int blocksize)
@@ -553,9 +615,9 @@ static inline int ocfs2_chain_recs_per_inode(int blocksize)
 	int size;
 
 	size = blocksize -
-		offsetof(struct _ocfs2_dinode, id2.i_chain.cl_recs);
+		offsetof(struct ocfs2_dinode, id2.i_chain.cl_recs);
 
-	return size / sizeof(struct _ocfs2_chain_rec);
+	return size / sizeof(struct ocfs2_chain_rec);
 }
 
 static inline int ocfs2_extent_recs_per_eb(int blocksize)
@@ -563,9 +625,9 @@ static inline int ocfs2_extent_recs_per_eb(int blocksize)
 	int size;
 
 	size = blocksize -
-		offsetof(struct _ocfs2_extent_block, h_list.l_recs);
+		offsetof(struct ocfs2_extent_block, h_list.l_recs);
 
-	return size / sizeof(struct _ocfs2_extent_rec);
+	return size / sizeof(struct ocfs2_extent_rec);
 }
 
 static inline int ocfs2_local_alloc_size(int blocksize)
@@ -573,7 +635,7 @@ static inline int ocfs2_local_alloc_size(int blocksize)
 	int size;
 
 	size = blocksize -
-		offsetof(struct _ocfs2_dinode, id2.i_lab.la_bitmap);
+		offsetof(struct ocfs2_dinode, id2.i_lab.la_bitmap);
 
 	return size;
 }
@@ -583,7 +645,7 @@ static inline int ocfs2_group_bitmap_size(int blocksize)
 	int size;
 
 	size = blocksize -
-		offsetof(struct _ocfs2_group_desc, bg_bitmap);
+		offsetof(struct ocfs2_group_desc, bg_bitmap);
 
 	return size;
 }
@@ -593,9 +655,22 @@ static inline int ocfs2_truncate_recs_per_inode(int blocksize)
 	int size;
 
 	size = blocksize -
-		offsetof(struct _ocfs2_dinode, id2.i_dealloc.tl_recs);
+		offsetof(struct ocfs2_dinode, id2.i_dealloc.tl_recs);
 
-	return size / sizeof(struct _ocfs2_truncate_rec);
+	return size / sizeof(struct ocfs2_truncate_rec);
+}
+
+static inline uint64_t ocfs2_backup_super_blkno(int blocksize, int index)
+{
+	uint64_t offset = OCFS2_BACKUP_SB_START;
+
+	if (index >= 0 && index < OCFS2_MAX_BACKUP_SUPERBLOCKS) {
+		offset <<= (2 * index);
+		offset /= blocksize;
+		return offset;
+	}
+
+	return 0;
 }
 #endif  /* __KERNEL__ */
 
