@@ -66,7 +66,7 @@ static void write_directory_data(State *s, DirData *dir);
 static void write_slot_map_data(State *s, SystemFileDiskRecord *slot_map_rec);
 static void write_group_data(State *s, AllocGroup *group);
 static void format_leading_space(State *s);
-static void replacement_journal_create(State *s, uint64_t journal_off);
+//static void replacement_journal_create(State *s, uint64_t journal_off);
 static void open_device(State *s);
 static void close_device(State *s);
 static int initial_slots_for_volume(uint64_t size);
@@ -83,6 +83,8 @@ static AllocGroup * initialize_alloc_group(State *s, const char *name,
 					   uint16_t chain, uint16_t cpg,
 					   uint16_t bpc);
 static void create_lost_found_dir(State *s);
+static void format_journals(State *s);
+static int format_backup_super(State *s);
 
 extern char *optarg;
 extern int optind, opterr, optopt;
@@ -101,6 +103,34 @@ static SystemFileInfo system_files[] = {
 	{ "truncate_log:%04d", SFI_TRUNCATE_LOG, 0, S_IFREG | 0644 }
 };
 
+struct fs_type_translation {
+	const char *ft_str;
+	enum ocfs2_fs_types ft_type;
+};
+
+static struct fs_type_translation ocfs2_fs_types_table[] = {
+	{"datafiles", FS_DATAFILES},
+	{"mail", FS_MAIL},
+	{NULL, FS_DEFAULT},
+};
+
+enum {
+	BACKUP_SUPER_OPTION = CHAR_MAX + 1,
+};
+
+static uint64_t align_bytes_to_clusters_ceil(State *s,
+					     uint64_t bytes)
+{
+	uint64_t ret = bytes + s->cluster_size - 1;
+
+	if (ret < bytes) /* deal with wrapping */
+		ret = UINT64_MAX;
+
+	ret = ret >> s->cluster_size_bits;
+	ret = ret << s->cluster_size_bits;
+
+	return ret;
+}
 
 static void
 handle_signal (int sig)
@@ -239,11 +269,6 @@ main(int argc, char **argv)
 	need = (s->volume_size_in_clusters + 7) >> 3;
 	need = ((need + s->cluster_size - 1) >> s->cluster_size_bits) << s->cluster_size_bits;
 
-	if (need > BITMAP_WARNING_LEN)
-		fprintf(stderr, "WARNING: bitmap is very large, consider using "
-				"a larger cluster size and/or\na smaller "
-				"volume\n");
-
 	if (!s->quiet)
 		printf("Creating bitmaps: ");
 
@@ -373,16 +398,6 @@ main(int argc, char **argv)
 		num = system_files[i].global ? 1 : s->initial_slots;
 		for (j = 0; j < num; j++) {
 			tmprec = &(record[i][j]);
-
-			if (system_files[i].type == SFI_JOURNAL) {
-				alloc_bytes_from_bitmap(s, s->journal_size_in_bytes,
-							s->global_bm,
-							&(tmprec->extent_off),
-							&(tmprec->extent_len));
-				replacement_journal_create(s, tmprec->extent_off);
-				tmprec->file_size = tmprec->extent_len;
-			}
-
 			format_file(s, tmprec);
 		}
 	}
@@ -428,7 +443,30 @@ main(int argc, char **argv)
 	if (!s->quiet)
 		printf("done\n");
 
+	if (!s->no_backup_super) {
+		if (!s->quiet)
+			printf("Writing backup superblock: ");
+
+		num = format_backup_super(s);
+		if (!s->quiet)
+			printf("%d block(s)\n", num);
+	}
+
 	if (!s->hb_dev) {
+		/* These routines use libocfs2 to do their work. We
+		 * don't share an ocfs2_filesys context between the
+		 * journal format and the lost+found create so that
+		 * the library can use the journal for the latter in
+		 * future revisions. */
+
+		if (!s->quiet)
+			printf("Formatting Journals: ");
+
+		format_journals(s);
+
+		if (!s->quiet)
+			printf("done\n");
+
 		if (!s->quiet)
 			printf("Writing lost+found: ");
 
@@ -446,6 +484,27 @@ main(int argc, char **argv)
 	return 0;
 }
 
+static void
+parse_fs_type_opts(char *progname, const char *typestr,
+		   enum ocfs2_fs_types *fs_type)
+{
+	int i;
+
+	*fs_type = FS_DEFAULT;
+
+	for(i = 0; ocfs2_fs_types_table[i].ft_str; i++) {
+		if (strcmp(typestr, ocfs2_fs_types_table[i].ft_str) == 0) {
+			*fs_type = ocfs2_fs_types_table[i].ft_type;
+			break;
+		}
+	}
+
+	if (*fs_type == FS_DEFAULT) {
+		com_err(progname, 0, "Bad fs type option specified.");
+		exit(1);
+	}
+}
+
 static State *
 get_state(int argc, char **argv)
 {
@@ -463,6 +522,9 @@ get_state(int argc, char **argv)
 	int ret;
 	uint64_t val;
 	uint64_t journal_size_in_bytes = 0;
+	enum ocfs2_fs_types fs_type = FS_DEFAULT;
+	int mount = 0;
+	int no_backup_super = 0;
 
 	static struct option long_options[] = {
 		{ "block-size", 1, 0, 'b' },
@@ -475,6 +537,8 @@ get_state(int argc, char **argv)
 		{ "journal-options", 0, 0, 'J'},
 		{ "heartbeat-device", 0, 0, 'H'},
 		{ "force", 0, 0, 'F'},
+		{ "mount", 1, 0, 'M'},
+		{ "no-backup-super", 0, 0, BACKUP_SUPER_OPTION },
 		{ 0, 0, 0, 0}
 	};
 
@@ -484,7 +548,7 @@ get_state(int argc, char **argv)
 		progname = strdup("mkfs.ocfs2");
 
 	while (1) {
-		c = getopt_long(argc, argv, "b:C:L:N:J:vqVFHx", long_options, 
+		c = getopt_long(argc, argv, "b:C:L:N:J:M:vqVFHxT:", long_options,
 				NULL);
 
 		if (c == -1)
@@ -540,6 +604,20 @@ get_state(int argc, char **argv)
 
 			break;
 
+		case 'M':
+			if (!strncasecmp(optarg, MOUNT_LOCAL_STR,
+					 strlen(MOUNT_LOCAL_STR)))
+				mount = MOUNT_LOCAL;
+			else if (!strncasecmp(optarg, MOUNT_CLUSTER_STR,
+					      strlen(MOUNT_CLUSTER_STR)))
+				mount = MOUNT_CLUSTER;
+			else {
+				com_err(progname, 0,
+					"Invalid mount type %s", optarg);
+				exit(1);
+			}
+			break;
+
 		case 'N':
 			initial_slots = strtoul(optarg, &dummy, 0);
 
@@ -587,6 +665,14 @@ get_state(int argc, char **argv)
 			xtool = 1;
 			break;
 
+		case 'T':
+			parse_fs_type_opts(progname, optarg, &fs_type);
+			break;
+
+		case BACKUP_SUPER_OPTION:
+			no_backup_super = 1;
+			break;
+
 		default:
 			usage(progname);
 			break;
@@ -647,6 +733,12 @@ get_state(int argc, char **argv)
 
 	s->hb_dev = hb_dev;
 
+	s->fs_type = fs_type;
+
+	s->mount = mount;
+
+	s->no_backup_super = no_backup_super;
+
 	return s;
 }
 
@@ -729,14 +821,11 @@ parse_journal_opts(char *progname, const char *opts,
 			ret = get_number(arg, &val);
 
 			if (ret ||
-			    val < OCFS2_MIN_JOURNAL_SIZE ||
-			    val > OCFS2_MAX_JOURNAL_SIZE) {
+			    val < OCFS2_MIN_JOURNAL_SIZE) {
 				com_err(progname, 0,
 					"Invalid journal size: %s\nSize must "
-					"be between %d and %d bytes",
-					arg,
-					OCFS2_MIN_JOURNAL_SIZE,
-					OCFS2_MAX_JOURNAL_SIZE);
+					"be greater than %d bytes",
+					arg, OCFS2_MIN_JOURNAL_SIZE);
 				exit(1);
 			}
 
@@ -759,11 +848,10 @@ parse_journal_opts(char *progname, const char *opts,
 static void
 usage(const char *progname)
 {
-	fprintf(stderr, "Usage: %s [-b block-size] [-C cluster-size] "
-			"[-N number-of-node-slots]\n"
-			"\t[-L volume-label] [-J journal-options] [-HFqvV] "
-			"device [blocks-count]\n",
-			progname);
+	fprintf(stderr, "usage: %s [-b block-size] [-C cluster-size] "
+		"[-J journal-options]\n\t\t[-L volume-label] [-M mount-type] "
+		"[-N number-of-node-slots]\n\t\t[-T filesystem-type] [-HFqvV] "
+		"[--no-backup-super] device [blocks-count]\n", progname);
 	exit(0);
 }
 
@@ -773,7 +861,41 @@ version(const char *progname)
 	fprintf(stderr, "%s %s\n", progname, VERSION);
 }
 
-/* stolen from e2fsprogs */
+static unsigned int journal_size_default(State *s)
+{
+	unsigned int j_blocks;
+
+	if (s->volume_size_in_blocks < 32768)
+		j_blocks = OCFS2_MIN_JOURNAL_SIZE / s->blocksize;
+	else if (s->volume_size_in_blocks < 262144)
+		j_blocks = 4096;
+	else {
+		/* Each journal gets ~.625% of the blocks in the file
+		 * system, with a min of 16384 and a max of 65536 */
+		j_blocks = s->volume_size_in_blocks / 160;
+		if (j_blocks < 16384)
+			j_blocks = 16384;
+		else if (j_blocks > 65536)
+			j_blocks = 65536;
+	}
+	return j_blocks;
+}
+
+static unsigned int journal_size_datafiles(void)
+{
+	return 8192;
+}
+
+static unsigned int journal_size_mail(State *s)
+{
+	if (s->volume_size_in_blocks < 262144)
+		return 16384;
+	else if (s->volume_size_in_blocks < 524288)
+		return 32768;
+	return 65536;
+}
+
+/* stolen from e2fsprogs, modified to fit ocfs2 patterns */
 static uint64_t figure_journal_size(uint64_t size, State *s)
 {
 	unsigned int j_blocks;
@@ -797,17 +919,70 @@ static uint64_t figure_journal_size(uint64_t size, State *s)
 				"Journal size too big for filesystem.\n");
 			exit(1);
 		}
-		return size;
+
+		return align_bytes_to_clusters_ceil(s, size);
 	}
 
-	if (s->volume_size_in_blocks < 32768)
-		j_blocks = OCFS2_MIN_JOURNAL_SIZE / s->blocksize;
-	else if (s->volume_size_in_blocks < 262144)
-		j_blocks = 4096;
-	else
-		j_blocks = 8192;
+	switch (s->fs_type) {
+	case FS_DATAFILES:
+		j_blocks = journal_size_datafiles();
+		break;
+	case FS_MAIL:
+		j_blocks = journal_size_mail(s);
+		break;
+	default:
+		j_blocks = journal_size_default(s);
+		break;
+	}
 
-	return j_blocks << s->blocksize_bits;
+	return align_bytes_to_clusters_ceil(s, j_blocks << s->blocksize_bits);
+}
+
+static uint32_t cluster_size_default(State *s)
+{
+	uint32_t volume_size, cluster_size, cluster_size_bits, need;
+
+	for (cluster_size = OCFS2_MIN_CLUSTERSIZE;
+	     cluster_size < AUTO_CLUSTERSIZE;
+	     cluster_size <<= 1) {
+		cluster_size_bits = get_bits(s, cluster_size);
+
+		volume_size =
+			s->volume_size_in_bytes >> cluster_size_bits;
+
+		need = (volume_size + 7) >> 3;
+		need = ((need + cluster_size - 1) >>
+			cluster_size_bits) << cluster_size_bits;
+
+		if (need <= BITMAP_AUTO_MAX) 
+			break;
+	}
+
+	return cluster_size;
+}
+
+static uint32_t cluster_size_datafiles(State *s)
+{
+	uint32_t cluster_size;
+	uint64_t volume_gigs = s->volume_size_in_bytes / (1024 * 1024 * 1024);
+
+	if (volume_gigs < 2) {
+		com_err(s->progname, 0,
+			"Selected file system type requires a device of at "
+			"least 2 gigabytes\n");
+		exit(1);
+	}
+
+	if (volume_gigs < 64)
+		cluster_size = 128;
+	else if (volume_gigs < 96)
+		cluster_size = 256;
+	else if (volume_gigs < 128)
+		cluster_size = 512;
+	else
+		cluster_size = 1024;
+
+	return cluster_size * 1024;
 }
 
 static void
@@ -940,32 +1115,23 @@ fill_defaults(State *s)
 	s->blocksize_bits = get_bits(s, s->blocksize);
 
 	if (!s->cluster_size) {
-		uint32_t volume_size, cluster_size, cluster_size_bits, need;
-
-		for (cluster_size = OCFS2_MIN_CLUSTERSIZE;
-		     cluster_size < AUTO_CLUSTERSIZE;
-		     cluster_size <<= 1) {
-			cluster_size_bits = get_bits(s, cluster_size);
-
-			volume_size =
-				s->volume_size_in_bytes >> cluster_size_bits;
-
-			need = (volume_size + 7) >> 3;
-			need = ((need + cluster_size - 1) >>
-				cluster_size_bits) << cluster_size_bits;
-
-			if (need <= BITMAP_AUTO_MAX) 
-				break;
+		switch (s->fs_type) {
+		case FS_DATAFILES:
+			s->cluster_size = cluster_size_datafiles(s);
+			break;
+		default:
+			s->cluster_size = cluster_size_default(s);
+			break;
 		}
-
-		s->cluster_size = cluster_size;
 	}
 
 	s->cluster_size_bits = get_bits(s, s->cluster_size);
 
+	/* volume size needs to be cluster aligned */
 	s->volume_size_in_clusters = s->volume_size_in_bytes >> s->cluster_size_bits;
 	tmp = (uint64_t)s->volume_size_in_clusters;
-	s->volume_size_in_blocks = (tmp << s->cluster_size_bits) >> s->blocksize_bits;
+	s->volume_size_in_bytes = tmp << s->cluster_size_bits;
+	s->volume_size_in_blocks = s->volume_size_in_bytes >> s->blocksize_bits;
 	
 	s->reserved_tail_size = 0;
 
@@ -982,8 +1148,11 @@ fill_defaults(State *s)
 	printf("tail_group_bits = %u\n", s->tail_group_bits);
 #endif
 	if (!s->initial_slots) {
-		s->initial_slots =
-			initial_slots_for_volume(s->volume_size_in_bytes);
+		if (s->mount == MOUNT_LOCAL)
+			s->initial_slots = 1;
+		else
+			s->initial_slots =
+				initial_slots_for_volume(s->volume_size_in_bytes);
 	}
 
 	if (!s->vol_label) {
@@ -1585,6 +1754,9 @@ format_superblock(State *s, SystemFileDiskRecord *rec,
 	if (s->hb_dev)
 		incompat |= OCFS2_FEATURE_INCOMPAT_HEARTBEAT_DEV;
 
+	if (s->mount == MOUNT_LOCAL)
+		incompat |= OCFS2_FEATURE_INCOMPAT_LOCAL_MOUNT;
+
 	di->id2.i_super.s_feature_incompat = incompat;
 
 	strcpy(di->id2.i_super.s_label, s->vol_label);
@@ -1854,45 +2026,6 @@ format_leading_space(State *s)
 	free(buf);
 }
 
-/*
- * The code to init a journal superblock is also in
- * libocfs2/mkjournal.c:ocfs2_init_journal_super_block().
- * Please keep them in sync.
- */
-static void
-replacement_journal_create(State *s, uint64_t journal_off)
-{
-	journal_superblock_t *sb;
-	void *buf;
-
-	buf = do_malloc(s, s->journal_size_in_bytes);
-	memset(buf, 0, s->journal_size_in_bytes);
-
-	sb = buf;
-
-	sb->s_header.h_magic     = JFS_MAGIC_NUMBER;
-	sb->s_header.h_blocktype = JFS_SUPERBLOCK_V2;
-
-	sb->s_blocksize = s->blocksize;
-	sb->s_maxlen = s->journal_size_in_bytes >> s->blocksize_bits;
-
-	if (s->blocksize == 512)
-		sb->s_first = 2;
-	else
-		sb->s_first = 1;
-
-	sb->s_start    = 1;
-	sb->s_sequence = 1;
-	sb->s_errno    = 0;
-	sb->s_nr_users = 1;
-
-	memcpy(sb->s_uuid, s->uuid, sizeof(sb->s_uuid));
-
-	ocfs2_swap_journal_superblock(sb);
-	do_pwrite(s, buf, s->journal_size_in_bytes, journal_off);
-	free(buf);
-}
-
 static void
 open_device(State *s)
 {
@@ -1997,9 +2130,20 @@ init_record(State *s, SystemFileDiskRecord *rec, int type, int mode)
 static void
 print_state(State *s)
 {
+	int i;
+
 	if (s->quiet)
 		return;
 
+	if (s->fs_type != FS_DEFAULT) {
+		for(i = 0; ocfs2_fs_types_table[i].ft_str; i++) {
+			if (ocfs2_fs_types_table[i].ft_type == s->fs_type) {
+				printf("Filesystem Type of %s\n",
+				       ocfs2_fs_types_table[i].ft_str);
+				break;
+			}
+		}
+	}
 	printf("Filesystem label=%s\n", s->vol_label);
 	printf("Block size=%u (bits=%u)\n", s->blocksize, s->blocksize_bits);
 	printf("Cluster size=%u (bits=%u)\n", s->cluster_size, s->cluster_size_bits);
@@ -2076,3 +2220,89 @@ bail:
 	clear_both_ends(s);
 	exit(1);
 }
+
+static void format_journals(State *s)
+{
+	errcode_t ret;
+	int i;
+	uint32_t journal_size_in_clusters;
+	uint64_t blkno;
+	ocfs2_filesys *fs = NULL;
+	char jrnl_file[40];
+
+	ret = ocfs2_open(s->device_name, OCFS2_FLAG_RW, 0, 0, &fs);
+	if (ret) {
+		com_err(s->progname, ret, "while opening new file system");
+		goto error;
+	}
+
+	journal_size_in_clusters = s->journal_size_in_bytes >>
+		OCFS2_RAW_SB(fs->fs_super)->s_clustersize_bits;
+
+	for(i = 0; i < OCFS2_RAW_SB(fs->fs_super)->s_max_slots; i++) {
+		snprintf (jrnl_file, sizeof(jrnl_file),
+			  ocfs2_system_inodes[JOURNAL_SYSTEM_INODE].si_name, i);
+		ret = ocfs2_lookup(fs, fs->fs_sysdir_blkno, jrnl_file,
+				   strlen(jrnl_file), NULL, &blkno);
+		if (ret) {
+			com_err(s->progname, ret,
+				"while looking up journal filename \"%.*s\"",
+				(int)strlen(jrnl_file), jrnl_file);
+			goto error;
+		}
+
+		ret = ocfs2_make_journal(fs, blkno, journal_size_in_clusters);
+		if (ret) {
+			com_err(s->progname, ret,
+				"while formatting journal \"%.*s\"",
+				(int)strlen(jrnl_file), jrnl_file);
+			goto error;
+		}
+	}
+
+	ocfs2_close(fs);
+	return;
+
+error:
+	clear_both_ends(s);
+	exit(1);
+}
+
+static int format_backup_super(State *s)
+{
+	errcode_t ret;
+	ocfs2_filesys *fs = NULL;
+	size_t len;
+	uint64_t blocks[OCFS2_MAX_BACKUP_SUPERBLOCKS];
+
+	ret = ocfs2_open(s->device_name, OCFS2_FLAG_RW, 0, 0, &fs);
+	if (ret) {
+		com_err(s->progname, ret,
+			"while opening file system for backup superblock.");
+		goto error;
+	}
+
+	len = ocfs2_get_backup_super_offset(fs, blocks, ARRAY_SIZE(blocks));
+
+	ret = ocfs2_set_backup_super(fs, blocks, len);
+	if (ret) {
+		com_err(s->progname, ret, "while backing up superblock.");
+		goto error;
+	}
+
+	OCFS2_SET_COMPAT_FEATURE(OCFS2_RAW_SB(fs->fs_super),
+				 OCFS2_FEATURE_COMPAT_BACKUP_SB);
+
+	ret = ocfs2_write_super(fs);
+	if (ret) {
+		com_err(s->progname, ret, "while updating superblock.");
+		goto error;
+	}
+
+	ocfs2_close(fs);
+	return len;
+
+error:
+	clear_both_ends(s);
+	exit(1);
+}
